from typing import Any, Dict, Optional

from ..db import Db
from ..deezer import Deezer
from ..sql import Exists, Insert
from ..utils import read_arl, url_to_deezer_id


class ChartSong:
    def __init__(self, data: Dict[str, Any]):
        self.title = data["title"]
        self.artist = data["artist"]
        self.label = data["label"]
        self.woc = data["woc"]  # Unused, can be unreliable from scraper, just calculate from db
        self.peak = data["peak"]
        self.position = data["position"]
        self.chart_id: int = data["chart_id"]
        self.deezer_id: str
        self.spotify_id: str
        self.track: Deezer = None
        self.track_id: Optional[int] = None

    def set_stream_urls(self, deezer_id: str, spotify_id: str) -> None:
        if deezer_id is not None:
            self.deezer_id = deezer_id
        if spotify_id is not None:
            self.spotify_id = spotify_id

    def to_deezer(self) -> Optional[Deezer]:
        try:
            initial_candidate = Deezer(
                read_arl("arl.txt"), url_to_deezer_id(self.deezer_id), spotify_url=self.spotify_id
            )
            if initial_candidate.album.track_count == 1:
                try:
                    song = initial_candidate.popular_similar()
                except ValueError:
                    song = initial_candidate
            else:
                song = initial_candidate

            self.track = song
            self.track_id = song.track_id
            return song
            # if initial_candidate.album.track_list
        except NameError:
            raise ValueError(f"No deezer id for song {self.title} by {self.artist}")

    def get_db_id(self, database: Db) -> int:
        if self.db_id is not None:
            return self.db_id
        else:
            return self.write_record(database)

    def write_record(self, database: Db, chart_db_id: int) -> int:
        self.chart_db_id = chart_db_id
        existing_id = database.exists(Exists.chart_track, (self.chart_db_id, self.track_id))
        if existing_id:
            self.db_id = existing_id
        else:
            database.insert(
                Insert.chart_track, (self.chart_db_id, self.track_id, self.position),
            )
            self.db_id = database.cursor.lastrowid
        return self.db_id


class ChartData:
    def __init__(self, data: Dict[str, Any]):
        self.chart_id = data["chart_id"]
        self.chart_title = data["chart_title"]
        self.chart_song: ChartSong = data["chart_song"]
        self.date = data["date"]
        self.chart_url = data["chart_url"]
        self.last_week = data["last_week"]
        self.db_id: Optional[int] = None

    def get_db_id(self, database: Db) -> int:
        if self.db_id is not None:
            return self.db_id
        else:
            return self.write_record(database)

    def write_record(self, database: Db) -> int:
        existing_id = database.exists(Exists.chart, (self.chart_title, self.chart_url, self.date))
        if existing_id:
            self.db_id = existing_id
        else:
            database.insert(Insert.chart, (self.chart_title, self.chart_url, self.date))
            self.db_id = database.cursor.lastrowid
        self.chart_song.write_record(database, self.get_db_id(database))
        return self.db_id



import time
from typing import Any, Optional

import acoustid
from mutagen.id3 import (
    APIC,
    RVA2,
    TALB,
    TBPM,
    TCON,
    TDEN,
    TDRL,
    TIT2,
    TPE1,
    TPE2,
    TPOS,
    TPUB,
    TRCK,
    TRSN,
    TSOA,
    TSOP,
    TSOT,
    TSRC,
    TXXX,
    USLT,
    WXXX,
)
from mutagen.mp3 import MP3

from .deezer import Deezer
from .mb import Musicbrainz
from .utils import sort_title


class Tag:
    def __init__(self, filename: str, deezer: Deezer, ext: str) -> None:
        self.filename = filename
        self.deezer = deezer
        self.ext = ext
        self._tag()

    def add_tag(self, f: MP3, tag: Any) -> None:
        try:
            f.tags.add(tag)
        except ValueError:
            return

    def _calculate_acoustid_fingerprint(self, filename: str) -> list:
        duration, fingerprint = acoustid.fingerprint_file(filename)

        return [duration, fingerprint]

    def _tag_mp3(self, filename: str, deezer: Deezer) -> None:
        f = MP3()

        f.add_tags()

        duration, acoustid_fingerprint = self._calculate_acoustid_fingerprint(filename)
        mb_ids = self._mb_id(deezer.contributor.name, deezer.track.title)

        self.add_tag(f, TPUB(encoding=3, text=deezer.album.label))
        self.add_tag(f, TSOT(encoding=3, text=deezer.track.title))
        self.add_tag(f, TSOA(encoding=3, text=sort_title(deezer.album.title)))
        self.add_tag(f, TSOP(encoding=3, text=deezer.contributor.name))
        self.add_tag(f, TDEN(text=time.strftime("%d/%m/%Y %H:%M:%S")))
        self.add_tag(f, TIT2(encoding=3, text=deezer.track.title))
        self.add_tag(f, TPOS(text=str(deezer.track.disk_number)))
        self.add_tag(f, TRCK(text=str(deezer.track.track_position)))
        self.add_tag(f, TDRL(text=deezer.album.id3v24_release_date))
        self.add_tag(f, TPE1(text=deezer.album.contributors))
        self.add_tag(f, TPE2(text=deezer.contributor.name))
        self.add_tag(f, TBPM(text=str(deezer.track.bpm)))
        self.add_tag(f, RVA2(desc="", channel=1, gain=deezer.track.gain))
        self.add_tag(f, USLT(text=deezer.lyrics.unsynced_lyrics()))
        self.add_tag(f, TCON(text=deezer.album.genre_list))
        self.add_tag(f, TXXX(text=deezer.album.upc, desc="upc"))
        self.add_tag(f, TALB(text=deezer.album.title))
        self.add_tag(f, WXXX(url=deezer.album.link, desc="album_deezer_url"))
        self.add_tag(f, WXXX(url=deezer.track.link, desc="song_deezer_url"))
        self.add_tag(f, WXXX(url=deezer.contributor.link, desc="artist_deezer_url"))
        self.add_tag(f, WXXX(url=deezer.track.spotify_url, desc="song_spotify_id"))
        self.add_tag(f, WXXX(url=acoustid_fingerprint, desc="acoust_id_fingerprint"))
        self.add_tag(
            f, APIC(type=3, data=deezer.album.cover.cover_art(1500), mime="image/jpeg")
        )  # 8 is artist, 3 is album
        self.add_tag(
            f, APIC(type=3, data=deezer.contributor.picture.cover_art(1500), mime="image/jpeg")
        )  # 8 is artist, 3 is album
        self.add_tag(f, TSRC(text=deezer.track.isrc))
        self.add_tag(f, TRSN(text="Deezer"))

        if mb_ids is not None:
            self.add_tag(f, TXXX(text=mb_ids["artist"], desc="musicbrainz_albumartistid"))
            self.add_tag(f, TXXX(text=mb_ids["album"], desc="musicbrainz_albumid"))
            self.add_tag(f, TXXX(text=mb_ids["quality"], desc="musicbrainz_id_quality"))

        f.save(filename)

    def _mb_id(self, artist: str, title: str) -> Optional[dict]:
        mb = Musicbrainz(artist, title)
        primary = mb.primary_mb_release

        if primary is None:
            return None

        try:
            ar_id = primary["artist-credit"][0]["artist"]["id"]
            al_id = primary["id"]
            qual = primary["mb_type"]
        except KeyError:  # There is somehow no artist or album id(?)
            return None

        data = {"artist": ar_id, "album": al_id, "quality": qual}
        return data

    def _tag(self) -> None:
        """ Function to write tags to the file, be it FLAC or MP3."""
        # retrieve tags

        if self.ext == ".mp3":
            self._tag_mp3(self.filename, self.deezer)
        else:
            raise NotImplementedError("Could not write tags. File extension not supported.")



import sqlite3
from typing import Optional, Tuple

from .sql import Create


class Db:
    def __init__(self, file_name: str) -> None:
        self.conn: sqlite3.Connection = sqlite3.connect(file_name)  # Creates/opens db
        self.conn.row_factory = sqlite3.Row
        self.cursor: sqlite3.Cursor = self.conn.cursor()
        self.create_tables()

    def commit(self) -> None:
        self.conn.commit()

    def select(self, sql: str, parameters: Tuple) -> None:
        self.cursor.execute(sql, parameters)

    def insert(self, sql: str, parameters: Tuple) -> None:
        self.cursor.execute(sql, parameters)

    def update(self, sql: str, parameters: Tuple) -> None:
        self.cursor.execute(sql, parameters)

    def exists(self, sql: str, parameters: Tuple) -> Optional[int]:
        data = self.cursor.execute(sql, parameters)
        record = data.fetchone()
        if record:
            return record[0]  # The 0th field is always the record ID
        else:
            return None

    def close(self) -> None:
        self.conn.close()

    def create_tables(self) -> None:
        self.cursor.execute(Create.chart)
        self.cursor.execute(Create.image_file)
        self.cursor.execute(Create.audio_file)
        self.cursor.execute(Create.photo)
        self.cursor.execute(Create.chart_track)
        self.cursor.execute(Create.contributor)
        self.cursor.execute(Create.album)
        self.cursor.execute(Create.genre)
        self.cursor.execute(Create.track)
        self.cursor.execute(Create.track_metadata)



from typing import Union, Dict


class Endpoints:
    @staticmethod
    def track(track_id: str) -> str:
        return f"https://api.deezer.com/track/{track_id}"

    @staticmethod
    def contributor(contributor_id: str) -> str:
        return f"https://api.deezer.com/artist/{contributor_id}"

    @staticmethod
    def album(album_id: str) -> str:
        return f"https://api.deezer.com/album/{album_id}"

    @staticmethod
    def genre(genre_id: str) -> str:
        return f"https://api.deezer.com/genre/{genre_id}"

    @staticmethod
    def search(parameters: Dict[str, Union[str, int]]) -> str:
        end_point = "https://api.deezer.com/search?q="
        for parameter, value in parameters.items():
            end_point += f"{parameter}:'{value}' "
        return end_point

    @staticmethod
    def get_user_data() -> str:
        return "deezer.getUserData"



from typing import Optional

import musicbrainzngs


class Musicbrainz:
    """ Class is used to make sure the release of the ID supplied as
        query_string is the main one.
    """

    def __init__(self, artist: str, title: str) -> None:
        musicbrainzngs.set_useragent("UK Chart Scraper", 1.0)

        self.artist = artist
        self.title = title
        self.mb_releases = self.search(self.artist, self.title)

        if self.mb_releases is None:
            self.primary_mb_release = None
        else:
            self.primary_mb_release = self.mb_releases[0]

    def search(self, artist: str, title: str) -> Optional[list]:
        possible_recordings = musicbrainzngs.search_recordings(
            query=f'title:"{title}" AND artistname:"{artist}" AND primarytype:"Album" AND status:"Official"', limit=100
        )["recording-list"]

        if len(possible_recordings) == 0:
            return None

        pristine_releases: list = []
        emergency_releases: list = []
        dire_releases: list = []

        for recording in possible_recordings:  # A recording is the sound made
            for release in recording["release-list"]:  # A release is the disc the sound is on
                release_type = self._classify_release(release)
                release["mb_type"] = release_type
                if release_type == "pristine":
                    if len(pristine_releases) == 0:
                        pristine_releases.append(release)
                    else:
                        pristine_releases = self._sort_countries(pristine_releases, release)
                elif release_type == "emergency":
                    # Just in case there are no perfect candidates
                    emergency_releases.append(release)
                elif release_type == "dire":
                    # Really bad releases that we really don't want
                    dire_releases.append(release)

        return self._prioritise_releases(pristine_releases, emergency_releases, dire_releases)

    def _prioritise_releases(self, pristine: list, emergency: list, dire: list) -> list:
        releases = []
        if len(pristine) == 0:
            if len(emergency) == 0:
                releases = dire
            else:
                releases = emergency
                releases.sort(key=lambda release: release["date"])
        else:
            releases = pristine

        return releases

    def _classify_release(self, release: dict) -> str:
        if (
            "secondary-type-list" not in release["release-group"]
            and "artist-credit-phrase" in release
            and release["artist-credit-phrase"] != "Various Artists"
            and "date" in release
        ):
            # Having a secondary-type-list means it's a compilation or something else unwanted
            return "pristine"
        elif "artist-credit-phrase" not in release and "date" in release:
            # Just in case there are no perfect candidates
            return "emergency"
        else:
            # Really bad releases that we really don't want
            return "dire"

    def _sort_countries(self, final_releases: list, release: dict) -> list:
        """ Inserts the release into the final_releases list favouring
            British releases, then European, then everywhere else.
        """

        if "country" not in release and "date" not in release:
            # Results like these do not have enough data to
            # be the best result.
            return final_releases
        if release["country"] == "GB":
            # Prefer UK releases
            release_list_index = 0
            while release["date"] > final_releases[release_list_index]["date"]:
                # Oldest UK releases first
                release_list_index += 1
                if release_list_index == len(final_releases):
                    break
            final_releases.insert(release_list_index, release)
        elif release["country"] == "XE":
            release_list_index = 0
            while final_releases[release_list_index]["country"] == "GB":
                # Keep going until all UK releases have been passed
                release_list_index += 1
            final_releases.insert(release_list_index, release)
        else:
            final_releases.append(release)

        return final_releases



class Insert:
    """Insert statements for all tables will eventually be included here
    """

    chart = """INSERT INTO "chart" (
            chart_id,
            name,
            url,
            date)
        VALUES (null, ?, ?, ?)
    """

    image_file = """INSERT INTO "image_file" (
            image_file_id,
            filepath,
            width,
            height,
            url)
        VALUES (null, ?, ?, ?, ?)
    """

    audio_file = """INSERT INTO "audio_file" (
            audio_file_id,
            filepath,
            format,
            bitrate,
            sample_rate,
            bits,
            channels,
            download_time)
        VALUES (null, ?, ?, ?, ?, ?, ?, ?)
    """

    photo = """INSERT INTO "photo" (
            photo_id,
            thumbnail_id,
            small_id,
            medium_id,
            big_id,
            xl_id)
        VALUES (null, ?, ?, ?, ?, ?)
    """

    chart_track = """INSERT INTO "chart_track" (
            chart_track_id,
            chart_id,
            track_id,
            position)
        VALUES (null, ?, ?, ?)
    """

    contributor = """INSERT INTO "contributor" (
            contributor_id,
            deezer_id,
            spotify_id,
            name,
            deezer_fans,
            photo_id)
        VALUES (null, ?, ?, ?, ?, ?)
    """

    album = """INSERT INTO "album" (
            album_id,
            deezer_id,
            spotify_id,
            title,
            primary_contributor,
            contributors,
            genre,
            duration,
            release_date,
            cover)
        VALUES (null, ?, ?, ?, ?, ?, ?, ?, ?, ?)
    """

    genre = """INSERT INTO "genre" (
            genre_id,
            name,
            photo)
        VALUES (null, ?, ?)
    """

    track = """INSERT INTO "track" (
            track_id,
            isrc,
            contributor_id,
            album_id,
            audio_file_id)
        VALUES (null, ?, ?, ?, ?)
    """

    track_metadata = """INSERT INTO "track_metadata" (
            isrc,
            deezer_id,
            spotify_id,
            title,
            short_title,
            duration,
            track_number,
            deezer_rank,
            explicit,
            bpm,
            lyrics)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
    """


class Update:
    """ Any update queries that can be performed on the database
    """

    set_track_audio_file = """
        UPDATE track
        SET audio_file_id = ?
        WHERE track_id = ?
    """


class Select:
    """Any needed select statements will be here
    """

    pass


class Exists:
    chart = """
        SELECT *
        FROM chart
        WHERE
            name = ? and
            url = ? and
            date = ?
    """

    image_file = """
        SELECT *
        FROM image_file
        WHERE filepath = ?
    """

    audio_file = """
        SELECT *
        FROM  audio_file
        WHERE filepath = ?
    """

    photo = """
        SELECT *
        FROM photo
        WHERE
            thumbnail_id = ? or
            small_id = ? or
            medium_id = ? or
            big_id = ? or
            xl_id = ?
    """

    chart_track = """
        SELECT *
        FROM chart_track
        WHERE
            chart_id = ? and
            track_id = ?
    """

    contributor = """
        SELECT *
        FROM contributor
        WHERE
            (deezer_id = ? or spotify_id = ?) and
            name = ?
    """

    album = """
        SELECT *
        FROM album
        WHERE
            (deezer_id = ? or spotify_id = ?) and
            title = ? and
            primary_contributor = ?
    """

    genre = """
        SELECT *
        FROM genre
        WHERE name = ?
    """

    track = """
        SELECT *
        FROM track
        WHERE isrc = ?
    """

    track_metadata = """
        SELECT *
        FROM track_metadata
        WHERE isrc = ?
    """


class Create:
    chart = """CREATE TABLE IF NOT EXISTS "chart" (
        "chart_id"    INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
        "name"    TEXT NOT NULL,
        "url"    TEXT NOT NULL,
        "date"    TIMESTAMP NOT NULL
    )"""

    image_file = """CREATE TABLE IF NOT EXISTS "image_file" (
        "image_file_id"    INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
        "filepath"    TEXT NOT NULL UNIQUE,
        "width"    INTEGER NOT NULL,
        "height"    INTEGER NOT NULL,
        "url"    INTEGER NOT NULL
    )"""

    audio_file = """CREATE TABLE IF NOT EXISTS "audio_file" (
        "audio_file_id"    INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
        "filepath"    TEXT NOT NULL UNIQUE,
        "format"    TEXT NOT NULL CHECK(format="mp3" or "flac"),
        "bitrate"    INTEGER NOT NULL,
        "sample_rate"    INTEGER NOT NULL,
        "bits"    INTEGER NOT NULL,
        "channels"    INTEGER NOT NULL,
        "download_time"    TIMESTAMP NOT NULL
    )"""

    photo = """CREATE TABLE IF NOT EXISTS "photo" (
        "photo_id"    INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
        "thumbnail_id"    INTEGER,
        "small_id"    INTEGER,
        "medium_id"    INTEGER,
        "big_id"    INTEGER,
        "xl_id"    INTEGER,
        FOREIGN KEY("xl_id") REFERENCES "image_file"("image_file_id"),
        FOREIGN KEY("big_id") REFERENCES "image_file"("image_file_id"),
        FOREIGN KEY("medium_id") REFERENCES "image_file"("image_file_id"),
        FOREIGN KEY("small_id") REFERENCES "image_file"("image_file_id"),
        FOREIGN KEY("thumbnail_id") REFERENCES "image_file"("image_file_id")
    )"""

    chart_track = """CREATE TABLE IF NOT EXISTS "chart_track" (
        "chart_track_id"    INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
        "chart_id"    INTEGER NOT NULL,
        "track_id"    INTEGER,
        "position"    INTEGER,
        FOREIGN KEY("chart_id") REFERENCES "chart"("chart_id"),
        FOREIGN KEY("track_id") REFERENCES "track"("track_id")
    )"""

    contributor = """CREATE TABLE IF NOT EXISTS "contributor" (
        "contributor_id"    INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
        "deezer_id"    INTEGER UNIQUE,
        "spotify_id"    TEXT UNIQUE,
        "name"    TEXT,
        "deezer_fans"    INTEGER,
        "photo_id"    INTEGER,
        FOREIGN KEY("photo_id") REFERENCES "photo"("photo_id")
    )"""

    album = """CREATE TABLE IF NOT EXISTS "album" (
        "album_id"    INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
        "deezer_id"    INTEGER UNIQUE,
        "spotify_id"    TEXT UNIQUE,
        "title"    TEXT,
        "primary_contributor"    INTEGER,
        "contributors"    TEXT,
        "genre"    INTEGER,
        "duration"    INTEGER,
        "release_date"    TIMESTAMP,
        "cover"    INTEGER,
        FOREIGN KEY("cover") REFERENCES "photo"("photo_id"),
        FOREIGN KEY("primary_contributor") REFERENCES "contributor"("contributor_id")
    )"""

    genre = """CREATE TABLE IF NOT EXISTS "genre" (
        "genre_id"    INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
        "name"    TEXT UNIQUE,
        "photo"    INTEGER,
        FOREIGN KEY("photo") REFERENCES "genre"("genre_id")
    )"""

    track = """CREATE TABLE IF NOT EXISTS "track" (
        "track_id"    INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
        "isrc"    TEXT NOT NULL UNIQUE,
        "contributor_id"    INTEGER NOT NULL,
        "album_id"    INTEGER NOT NULL,
        "audio_file_id"    INTEGER,
        FOREIGN KEY("contributor_id") REFERENCES "contributor"("contributor_id"),
        FOREIGN KEY("album_id") REFERENCES "album"("album_id"),
        FOREIGN KEY("isrc") REFERENCES "track_metadata"("isrc")
    )"""

    track_metadata = """CREATE TABLE IF NOT EXISTS "track_metadata" (
        "isrc"    TEXT NOT NULL UNIQUE,
        "deezer_id"    INTEGER,
        "spotify_id"    INTEGER,
        "title"    TEXT,
        "short_title"    TEXT,
        "duration"    INTEGER,
        "track_number"    INTEGER,
        "deezer_rank"    INTEGER,
        "explicit"    INTEGER NOT NULL CHECK(explicit = 1 or explicit = 0),
        "bpm"    REAL,
        "lyrics"    TEXT,
        PRIMARY KEY("isrc")
    )"""



from typing import Dict, Optional, cast

import requests

from .db import Db
from .endpoints import Endpoints
from .models import deezer
from .sql import Exists, Insert
from .utils import replace_illegals, requests_retry_session


class Deezer:
    # Be careful when debugging this as VSCode's sidebar will call the @property methods
    def __init__(self, arl: str, track_id: str, spotify_url: str = None) -> None:
        self.initialise_counter = 0
        self.initialise(arl, track_id, spotify_url)

    def initialise(self, arl: str, track_id: str, spotify_url: str = None) -> None:
        self.initialise_counter += 1
        self.arl = arl
        self._setup_session()
        self._login_by_arl(self.arl)
        self._get_tokens()
        self.track_id = track_id

        self.lyrics = Lyrics(track_id, self)

        self._raw_track_data: dict = self._request_wrapper(Endpoints.track(track_id))

        self._track_data: deezer.Track = None
        self._album_data: deezer.Album = None
        self._artist_data: deezer.Contributor = None
        self.spotify_url = spotify_url

        self.artist_id = self._raw_track_data["artist"]["id"]
        self.album_id = self._raw_track_data["album"]["id"]

    def _setup_session(self) -> None:
        self.session = requests.Session()
        user_agent = (
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/68.0.3440.106 Safari/537.36"
        )
        http_headers = {
            "User-Agent": user_agent,
            "Content-Language": "en-US",
            "Cache-Control": "max-age=0",
            "Accept": "*/*",
            "Accept-Charset": "utf-8,ISO-8859-1;q=0.7,*;q=0.3",
            "Accept-Language": "en-US;q=0.6,en;q=0.4",
            "Connection": "keep-alive",
        }
        self.session.headers.update(http_headers)

    def _login_by_arl(self, arl: str) -> bool:
        cookies = {"arl": arl}
        self.session.cookies.update(cookies)
        req = self._perform_hidden_request(Endpoints.get_user_data())
        if req["USER"]["USER_ID"]:
            return True
        else:
            return False

    def _perform_hidden_request(self, method: str, json_req: dict = {}) -> dict:
        """ Requests info from the hidden api: gw-light.php.
            Used for loginUserToken(), getCSRFToken()
            and privateApi().
        """
        api_token = "null" if method == Endpoints.get_user_data() else self.csrf_token
        unofficial_api_queries = {
            "api_version": "1.0",
            "api_token": api_token,
            "input": "3",
            "method": method,
        }
        req = (
            requests_retry_session(session=self.session)
            .post(url="https://www.deezer.com/ajax/gw-light.php", params=unofficial_api_queries, json=json_req,)
            .json()
        )
        return req["results"]

    def _get_tokens(self) -> None:
        req = self._perform_hidden_request(Endpoints.get_user_data())
        self.csrf_token = req["checkForm"]
        self.sid_token = req["SESSION_ID"]

    def _perform_request(self, end_point: str) -> Optional[Dict]:
        resp = requests_retry_session().get(end_point)
        data = resp.json()

        if "error" in data:
            return None

        return data

    def _request_wrapper(self, end_point: str) -> Optional[Dict]:
        data = self._perform_request(end_point)
        if not data:
            if self.initialise_counter == 3:
                raise ValueError(f"No deezer id for song {self.track_id}")
            search_end_point = Endpoints.search(
                {"track": self._get_track(children=False).title, "artist": self.contributor.name}
            )
            search_data = self._perform_request(search_end_point)
            self.initialise(self.arl, search_data["data"][0]["id"], self.spotify_url)
            data = self._request_wrapper(end_point)

            return data

        return data

    def popular_similar(self) -> "Deezer":
        # Use a string because Deezer isn't defined when Deezer is being defined.
        search_end_point = Endpoints.search({"track": self.track.title, "artist": self.contributor.name})
        data = self._request_wrapper(search_end_point)
        return Deezer(self.arl, data["data"][0]["id"], self.spotify_url)

    def _get_track(self, children: bool = True) -> deezer.Track:
        if self._track_data is None:
            if children:
                self._artist_data = self.contributor
                self._album_data = self.album

                data = self._raw_track_data
                data["artist"] = self._artist_data
                data["album"] = self._album_data
                self._track_data = deezer.Track(data, self.spotify_url, net_req=True)
            else:
                data = self._raw_track_data
                data["artist"] = {}
                data["album"] = {}

                self._track_data = deezer.Track(data, self.spotify_url, net_req=True, skip_children=True)
        return self._track_data

    track: deezer.Track = property(_get_track)

    @property
    def album(self) -> deezer.Album:
        if self._album_data is None:
            raw_album_data = self._request_wrapper(Endpoints.album(self.album_id))
            self._album_data = deezer.Album(raw_album_data, net_req=True)

        return self._album_data

    @property
    def contributor(self) -> deezer.Contributor:
        if self._artist_data is None:
            raw_artist_data = self._request_wrapper(Endpoints.contributor(self.artist_id))
            self._artist_data = deezer.Contributor(raw_artist_data, net_req=True)

        return self._artist_data

    def generate_filepath(self, prefix: str) -> str:
        filepath = f"{prefix}/{self.contributor.name}/{replace_illegals(self.album.title)}/{replace_illegals(self.track.title_short)}"
        return filepath

    def get_lyrics(self, synced: bool = False) -> list:
        if synced:
            return self.lyrics.synced_lyrics()
        else:
            return self.lyrics.unsynced_lyrics()

    def write_record(self, database: Db) -> None:
        self.track.write_record(database)
        self.write_metadata(database)

    def write_metadata(self, database: Db) -> int:
        existing_id = database.exists(Exists.track_metadata, (self.track.isrc,))
        if existing_id:
            db_id = existing_id
        else:
            database.insert(
                Insert.track_metadata,
                (
                    self.track.isrc,
                    self.track_id,
                    self.track.spotify_url,
                    self.track.title,
                    self.track.title_short,
                    self.track.duration,
                    self.track.track_position,
                    self.track.rank,
                    self.track.explicit_lyrics,
                    self.track.bpm,
                    "\n".join(self.lyrics.unsynced_lyrics()),
                ),
            )
            db_id = cast(int, database.cursor.lastrowid)
        return db_id


class Lyrics:
    def __init__(self, track_id: str, deezer: Deezer):
        self.track_id = track_id
        self.deezer = deezer
        self._lyric_data: dict = {}

    def _lyrics_request(self) -> dict:
        self._lyric_data = self.deezer._perform_hidden_request("song.getLyrics", {"sng_id": self.track_id})

        return self._lyric_data

    def unsynced_lyrics(self) -> list:
        if self._lyric_data == {}:
            data = self._lyrics_request()
        else:
            data = self._lyric_data
        if "LYRICS_TEXT" in data:  # unsynced lyrics
            lyrics = data["LYRICS_TEXT"].splitlines()  # True keeps the \n
            return lyrics
        else:
            return []

    def synced_lyrics(self) -> list:
        if self._lyric_data == {}:
            data = self._lyrics_request()
        else:
            data = self._lyric_data
        if "LYRICS_SYNC_JSON" in data:  # synced lyrics
            lyrics = data["LYRICS_SYNC_JSON"]
            return lyrics
        else:
            return []



import datetime
from typing import Dict, Generator, List, Optional, Tuple, Union

import bs4
import requests
from bs4 import BeautifulSoup as bs

from .models import chart


class Scraper:
    def __init__(self, date: datetime.date, chart_title: str):
        self.date: datetime.date = date
        self.chart_title: str = chart_title
        self.chart_id = self._chart_id_from_type(self.chart_title)
        self.chart_date = date.strftime("%Y%m%d")
        self.chart_url = self._chart_url(self.chart_title, self.chart_date, self.chart_id)

    def scrape(self) -> Generator[chart.ChartData, None, None]:
        soup: bs = self._download_webpage(self.chart_url)

        rows: List[bs4.element.Tag]
        data_table: bs4.element.Tag
        rows, data_table = self._extract_rows(soup)
        for row in rows:
            if self._is_advertisement(row):
                continue

            chart_song = self._extract_chartsong_metadata(row)
            stream_links = self._extract_stream_links(row, data_table)

            for link in stream_links:
                provider = self._streaming_provider(link)
                if provider == "deezer":
                    deezer_stream_url = link
                elif provider == "spotify":
                    spotify_stream_url = link

            chart_song.set_stream_urls(deezer_stream_url, spotify_stream_url)

            chart_data = self._extract_chartdata_metadata(row, chart_song)
            yield chart_data

    def _chart_id_from_type(self, chart_id: str) -> str:
        chart_ids = {"singles-chart": "7501"}
        return chart_ids[chart_id]

    def _is_advertisement(self, row: bs4.element.Tag) -> bool:

        if row.find(class_="adspace"):
            return True
        else:
            return False

    def _chart_url(self, chart_type: str, chart_date: str, chart_id: str) -> str:
        chart_url = f"http://www.officialcharts.com/charts/{chart_type}/{chart_date}/{chart_id}/"

        return chart_url

    def _download_webpage(self, url: str) -> bs:
        page = requests.get(url)  # DLs the chart page
        soup = bs(page.content, "html.parser")  # Initialises into BS4

        return soup

    def _extract_rows(self, parsed_page: bs) -> Tuple[List[bs4.element.Tag], bs4.element.Tag]:
        data_table = parsed_page.find("table", class_="chart-positions")

        data_rows = data_table.find_all(
            "tr", class_=None
        )  # Creates a list of all the "Rows" of the tablespecial type of row

        return data_rows, data_table

    def _extract_stream_links(self, row: bs4.element.Tag, parsed_page: bs) -> List[str]:
        """ Finds all of the rows with streaming links in.
            Each of these has an index as part of their ID, get this.
        """
        try:
            # Gets all of the links, we only want the first (0th) one
            stream_open_button = row.find_all("div", class_="actions")[0].find_all("a", class_="")[0]

            # Splits the 'data-toggle' id into the hyphenated parts. Only the end part is unique
            stream_open_button_id: int = int(stream_open_button["data-toggle"].split("-")[3])
        except IndexError:
            # The open button doesn't exist in the row,
            # so there are no streaming links
            return []

        stream_elements = parsed_page.find_all("tr", class_=f"actions-view-listen-{stream_open_button_id}")
        stream_links: List[str] = [tag.get("href") for tag in stream_elements[0].find_all("a")]
        # Hopefully, if both Deezer and Spotify stream the song, a list of two links

        return stream_links

    def _streaming_provider(self, link: str) -> Optional[str]:
        if "deezer" in link:
            return "deezer"
        elif "spotify" in link:
            return "spotify"
        else:
            return None

    def _add_stream_links_to_object(self, chart_data: chart.ChartData, deezer: str, spotify: str) -> None:
        chart_data.set_stream_urls(deezer, spotify)

    def _extract_chartsong_metadata(self, row: bs4.element.Tag) -> chart.ChartSong:
        row_data: Dict[str, Union[str, int]] = {}

        row_data["title"] = row.find(class_="title").find("a").get_text()
        row_data["artist"] = row.find(class_="artist").find("a").get_text()
        row_data["label"] = row.find(class_="label").get_text()
        row_data["peak"] = row.find_all("td", recursive=False)[3].get_text()
        row_data["woc"] = row.find_all("td", recursive=False)[4].get_text()
        row_data["position"] = row.find(class_="position").get_text()
        row_data["chart_id"] = self.chart_id

        return chart.ChartSong(row_data)

    def _extract_chartdata_metadata(self, row: bs4.element.Tag, chart_song: chart.ChartSong) -> chart.ChartData:
        # Make some way to have a ChartSong in here in the parameters
        row_data: Dict[str, Union[str, int]] = {}

        row_data["chart_id"] = self.chart_id
        row_data["chart_title"] = self.chart_title
        row_data["chart_song"] = chart_song
        row_data["date"] = self.chart_date
        row_data["chart_url"] = self.chart_url
        row_data["last_week"] = row.find(class_="last-week").get_text().strip()

        # Make sure these corruspond with the ones in ChartData

        return chart.ChartData(row_data)



import base64
from typing import Dict

import requests


class Spotify:
    def __init__(self, client_id: str, client_secret: str) -> None:
        self.client_id = client_id
        self.client_secret = client_secret
        self.headers = self.spotify_headers(self.client_id, self.client_id)

    def spotify_headers(self, client_id: str, client_secret: str) -> Dict:
        data = {"grant_type": "client_credentials"}
        auth_code = base64.b64encode(f"{client_id}:{client_secret}".encode()).decode()
        headers = {"Authorization": f"Basic {auth_code}"}
        resp = requests.post("https://accounts.spotify.com/api/token", data=data, headers=headers)

        auth_token = resp.json()["access_token"]

        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": f"Bearer {auth_token}",
        }

        return headers

    def refresh_headers(self) -> None:
        self.headers = self.spotify_headers(self.client_id, self.client_secret)



import datetime
from typing import Any, Dict, List, Optional, cast

from ..db import Db
from ..endpoints import Endpoints
from ..sql import Exists, Insert
from ..utils import dict_rename, requests_retry_session


class Track:
    def __init__(
        self, data: Dict[str, Any], spotify_url: str = None, net_req: bool = False, skip_children: bool = False
    ):
        if net_req:
            data = self._net_req_convert(data, skip_children)

        self.deezer_id: str = data["deezer_id"]
        self.title: str = data["title"]
        self.title_short: str = data["title_short"]
        self.isrc: str = data["isrc"]
        self.link: str = data["link"]
        self.duration: int = data["duration"]
        self.track_position: int = data["track_position"]
        self.disk_number: int = data["disk_number"]
        self.rank: int = data["rank"]
        self.release_date: datetime.date = data["release_date"]
        self.explicit_lyrics: bool = data["explicit_lyrics"]
        self.preview_url: str = data["preview_url"]
        self.bpm: int = data["bpm"]
        self.gain: float = data["gain"]
        self.contributors: List[int] = data["contributors"]
        self.artist: Contributor = data["artist"]
        self.album: Album = data["album"]
        self.spotify_url: Optional[str] = spotify_url

        if "title_version" in data:
            self.title_version: str = data["title_version"]

        self.db_id: Optional[int] = None

    def _net_req_convert(self, data: Dict[str, Any], skip_children: bool = False) -> Dict[str, Any]:
        rename = {"id": "deezer_id", "preview": "preview_url"}
        remove = [
            "share",
            "explicit_content_lyrics",
            "explicit_content_cover",
            "available_countries",
        ]

        for original, new in rename.items():
            data[new] = data[original]
            data.pop(original)

        for field in remove:
            data.pop(field)

        data["contributors"] = [contrib_data["id"] for contrib_data in data["contributors"]]

        if not skip_children:
            if not isinstance(data["artist"], Contributor):
                data["artist"] = Contributor(data["artist"])
            if not isinstance(data["album"], Album):
                data["album"] = Album(data["album"])

        return data

    def get_db_id(self, database: Db) -> int:
        if self.db_id is not None:
            return self.db_id
        else:
            return self.write_record(database)

    def write_record(self, database: Db) -> int:
        existing_id = database.exists(Exists.track, (self.isrc,))
        if existing_id:
            self.db_id = existing_id
        else:
            database.insert(
                Insert.track, (self.isrc, self.artist.get_db_id(database), self.album.get_db_id(database), None)
            )
            self.db_id = cast(int, database.cursor.lastrowid)
        return self.db_id


class Contributor:
    def __init__(self, data: Dict[str, Any], net_req: bool = False):
        if net_req:
            data = self._net_req_convert(data)
        self.id: int = data["id"]
        self.name: str = data["name"]
        self.link: str = data["link"]
        self.picture: Picture = data["picture"]
        self.albums: int = data["albums"]
        self.fans: int = data["fans"]
        self.track_list: str = data["track_list"]

        self.db_id: Optional[int] = None

    def _net_req_convert(self, data: Dict[str, Any]) -> Dict[str, Any]:
        removed = ["share", "radio", "type"]
        renamed = {"nb_album": "albums", "nb_fan": "fans", "tracklist": "track_list"}

        for field in removed:
            data.pop(field)

        for original, new in renamed.items():
            data[new] = data[original]
            data.pop(original)

        data["picture"] = Picture(data, net_req=True)

        return data

    def get_db_id(self, database: Db) -> int:
        if self.db_id is not None:
            return self.db_id
        else:
            return self.write_record(database)

    def write_record(self, database: Db) -> int:
        existing_id = database.exists(Exists.contributor, (self.id, None, self.name))
        if existing_id:
            self.db_id = existing_id
        else:
            database.insert(Insert.contributor, (self.id, None, self.name, self.fans, self.picture.get_db_id(database)))
            self.db_id = cast(int, database.cursor.lastrowid)
        return self.db_id


class Picture:
    def __init__(self, data: Dict[str, Any], net_req: bool = False, prefix: str = "picture"):
        if net_req:
            data = self._net_req_convert(data, prefix)

        self.thumbnail: dict = {}
        self.small: dict = {}
        self.medium: dict = {}
        self.big: dict = {}
        self.xl: dict = {}

        self.thumbnail["size"] = [120, 120]

        self.thumbnail["url"] = data["thumbnail"]

        if all(size in data.keys() for size in ["small", "medium", "big", "xl"]):
            self.small["size"] = [56, 56]
            self.small["url"] = data["small"]

            self.medium["url"] = data["medium"]
            self.medium["size"] = [250, 250]

            self.big["url"] = data["big"]
            self.big["size"] = [500, 500]

            self.xl["url"] = data["xl"]
            self.xl["size"] = [1000, 1000]

        self.db_id: Optional[int] = None

    def _net_req_convert(self, data: Dict[str, Any], prefix: str = "picture") -> Dict[str, Any]:
        keys = {
            f"{prefix}": "thumbnail",
            f"{prefix}_small": "small",
            f"{prefix}_medium": "medium",
            f"{prefix}_big": "big",
            f"{prefix}_xl": "xl",
        }

        picture_data = {}

        try:
            for original, new in keys.items():
                picture_data[new] = data[original]
        except KeyError:  # Only a single image
            return picture_data
        return picture_data

    @property
    def _url_prefix(self) -> str:
        return "/".join(self.xl["url"].split("/")[0:-1])

    def cover_art(self, size: int) -> bytes:
        """ Retrieves the cover art/playlist image from the official API,
            downloads and returns it.
        """
        url = f"{self._url_prefix}/{size}x{size}.png"
        r = requests_retry_session().get(url)
        return r.content

    def get_db_id(self, database: Db) -> int:
        if self.db_id is not None:
            return self.db_id
        else:
            return self.write_record(database)

    def _write_image_file_record(self, database: Db, image: dict) -> int:
        existing_id = database.exists(Exists.image_file, (image["url"],))
        if existing_id:
            return existing_id
        else:
            database.insert(Insert.image_file, (image["url"], image["size"][0], image["size"][1], True))
            return database.cursor.lastrowid

    def write_image_file(self, database: Db) -> None:
        if all([self.xl == {}, self.big == {}, self.medium == {}, self.small == {}]):
            images = [self.thumbnail]
        else:
            images = [self.thumbnail, self.small, self.medium, self.big, self.xl]
        for i, image in enumerate(images):
            images[i]["db_id"] = self._write_image_file_record(database, image)

    def write_record(self, database: Db) -> int:
        if all([self.xl == {}, self.big == {}, self.medium == {}, self.small == {}]):
            image_ids = (self._write_image_file_record(database, self.thumbnail), None, None, None, None)
            existing_id = database.exists(Exists.photo, image_ids)
        else:
            images = [self.thumbnail, self.small, self.medium, self.big, self.xl]
            image_ids = tuple(self._write_image_file_record(database, image) for image in images)
            existing_id = database.exists(Exists.photo, image_ids)
        if existing_id:
            self.db_id = existing_id
        else:
            self.write_image_file(database)
            database.insert(Insert.photo, image_ids)
            self.db_id = cast(int, database.cursor.lastrowid)
        return self.db_id


class Album:
    def __init__(self, data: Dict[str, Any], net_req: bool = False):
        if net_req:
            data = self._net_req_convert(data)

        self.id: int = data["id"]
        self.title: str = data["title"]
        self.upc: str = data["upc"]
        self.link: str = data["link"]
        self.cover: Picture = data["cover"]
        self.genres: List[Genre] = data["genres"]
        self.label: str = data["label"]
        self.duration: int = data["duration"]
        self.fans: int = data["fans"]
        self.rating: int = data["rating"]
        self.release_date: datetime.date = data["release_date"]
        self.record_type = data["record_type"]
        self.track_list: str = data["track_list"]
        self.explicit_lyrics: bool = data["explicit_lyrics"]
        self.explicit_content_lyrics: bool = data["explicit_content_lyrics"]
        self.explicit_content_cover: bool = data["explicit_content_cover"]
        self.contributors: List[int] = data["contributors"]
        self.artist: Contributor = data["artist"]
        self.track_count: int = data["track_count"]

        self.db_id: Optional[int] = None

    @property
    def id3v24_release_date(self) -> str:
        return self.release_date.strftime("%Y%m%d")

    @property
    def genre_list(self) -> List[str]:
        genres = [genre.name for genre in self.genres]
        return genres

    def _net_req_convert(self, data: Dict[str, Any]) -> Dict[str, Any]:

        rename = {"tracklist": "track_list", "nb_tracks": "track_count"}
        cover_keys = ["cover", "cover_small", "cover_medium", "cover_big", "cover_xl"]

        data["contributors"] = [contrib_data["id"] for contrib_data in data["contributors"]]
        data["genres"] = [Genre(genre_data, net_req=True) for genre_data in data["genres"]["data"]]
        data["release_date"] = datetime.datetime.strptime(data["release_date"], "%Y-%m-%d")

        # Note, this always runs, even when there already is complete data.
        artist_data = self._perform_request(Endpoints.contributor(data["artist"]["id"]))

        data["artist"] = Contributor(artist_data, net_req=True)

        cover_data = {}
        for key in cover_keys:
            cover_data[key] = data[key]

        data["cover"] = Picture(cover_data, net_req=True, prefix="cover")

        data = dict_rename(data, rename)

        # Means that a Track() doesn't reference an Album() that contains the same Track()
        data.pop("tracks")

        return data

    def _perform_request(self, end_point: str) -> Optional[Dict]:
        resp = requests_retry_session().get(end_point)
        data = resp.json()

        if "error" in data:
            return None

        return data

    def get_db_id(self, database: Db) -> int:
        if self.db_id is not None:
            return self.db_id
        else:
            return self.write_record(database)

    def write_record(self, database: Db) -> int:
        existing_id = database.exists(Exists.album, (self.id, None, self.title, self.artist.get_db_id(database)))
        if existing_id:
            self.db_id = existing_id
        else:
            genre_db_ids = [genre.get_db_id(database) for genre in self.genres]
            database.insert(
                Insert.album,
                (
                    self.id,
                    None,
                    self.title,
                    self.artist.get_db_id(database),
                    str(self.contributors),
                    str(genre_db_ids),
                    self.duration,
                    self.release_date,
                    self.cover.get_db_id(database),
                ),
            )
            self.db_id = cast(int, database.cursor.lastrowid)
        return self.db_id


class Genre:
    def __init__(self, data: Dict[str, Any], net_req: bool = False):
        if net_req:
            data = self._net_req_convert(data)

        self.id: int = data["id"]
        self.name: str = data["name"]
        self.picture: Picture = data["picture"]

        self.db_id: Optional[int] = None

    def _net_req_convert(self, data: dict) -> dict:
        genre_data = self._perform_request(Endpoints.genre(str(data["id"])))
        genre_data = cast(dict, genre_data)

        data["picture"] = Picture(genre_data, net_req=True, prefix="picture")

        return data

    def _perform_request(self, end_point: str) -> Optional[Dict]:
        resp = requests_retry_session().get(end_point)
        data = resp.json()

        if "error" in data:
            return None

        return data

    def get_db_id(self, database: Db) -> int:
        if self.db_id is not None:
            return self.db_id
        else:
            return self.write_record(database)

    def write_record(self, database: Db) -> int:
        existing_id = database.exists(Exists.genre, (self.name,))
        if existing_id:
            self.db_id = existing_id
        else:
            database.insert(Insert.genre, (self.name, self.picture.get_db_id(database)))
            self.db_id = cast(int, database.cursor.lastrowid)
        return self.db_id



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="uk_charts_api",
    version="1.0.3",
    author="Connor McLaughlin",
    description="Python package to get deezer info for UK Top 40 charts",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cmclaughlin/UK-Charts-API",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Multimedia :: Sound/Audio",
    ],
    python_requires=">=3.6",
)



from datetime import date, timedelta
from typing import Any, Dict, List, Optional, Tuple

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry


def generate_dates(year: int) -> List[date]:
    """[Generates a list of all the week beginings in year. Note: Weeks start on Sundays]

    Arguments:
        year {int} -- [The year to generate a list of dates for]

    Returns:
        List[str] -- [List of str representations of datetime.date objects for the year]
    """
    dates = []
    d = date(year, 1, 1)  # January 1st
    d += timedelta(days=6 - d.weekday())  # First Sunday
    while d.year == year:
        dates.append(d)
        d += timedelta(days=7)
        if d.year != year:
            break
    return dates


def dict_rename(data: Dict[str, Any], rename: Dict[str, Any]) -> Dict[str, Any]:
    for original, new in rename.items():
        data[new] = data[original]
        data.pop(original)

    return data


def read_arl(filename: str) -> str:
    with open(filename) as file:
        arl = file.readlines()[0].strip()
        return arl


def read_acoustid(filename: str) -> str:
    with open(filename) as file:
        key = file.readlines()[0].strip()
        return key


def requests_retry_session(
    retries: int = 3,
    backoff_factor: float = 0.3,
    status_forcelist: Tuple[int, int, int] = (500, 502, 504),
    session: Optional[requests.Session] = None,
) -> requests.Session:
    session = session or requests.Session()
    retry = Retry(
        total=retries, read=retries, connect=retries, backoff_factor=backoff_factor, status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


def sort_title(text: str) -> str:
    text.lstrip("A ").lstrip("The ")

    return text


def url_to_deezer_id(url: str) -> str:
    return url.split("/")[-1]


def replace_illegals(text: str) -> str:
    text.replace("/", "_")
    return text
